@@ -22,12 +22,16 @@ def create_interpolator(method, values, F, M, S, npy=None,degree=-1, log_level=0
     intr = None
     if method == "linear":
         intr = Multivariate(values,F, M, S,1, log_level)
+    ## DEGREE 5 explicitly ##
     elif method == "cubic" and degree == 5: #not really cubic, but quintic
         intr = Multivariate(values,F, M, S,5, log_level)
-    elif method == "cubic":
-        intr = Multivariate(values,F, M, S,3, log_level)
     elif method == "bspline" and degree == 5:
         intr = Bspline(values,F, M, S, 5, log_level) # I am being cautious about degrees
+    ### degree 3 ###
+    elif method == "cubic":
+        intr = Multivariate(values,F, M, S,3, log_level)    
+    elif method == "mspline":
+        intr = Mspline(values,F, M, S,3, log_level) 
     elif method == "bspline":
         intr = Bspline(values,F, M, S,3, log_level) 
     else: #nearest is default
@@ -38,14 +42,17 @@ def create_interpolator(method, values, F, M, S, npy=None,degree=-1, log_level=0
 ### Abstract class ############################################################################
 class Interpolator(ABC):
     def __init__(self, values, F, M, S, degree=-1,log_level=0):                             
-        self._npy = np.copy(values)
+        self._npy = np.copy(values).astype(float)
         self._F = F
         self._M = M
         self._S = S
-        self.degree = degree
-        self._buffer = 12
+        self.degree = degree        
         self.log_level = log_level         
-        self.h = 0.01 #this is the interval for numerical differentiation
+        self.h = 0.0001 #this is the interval for numerical differentiation
+        self.padded = False
+        self.buffer = 14
+        self.round = 12
+        self.TOLERANCE = 2.2204460492503131e-016 # smallest such that 1.0+DBL_EPSILON != 1.0                                
     ############################################################################################################            
     @abstractmethod
     def get_value(self, x, y, z):
@@ -59,22 +66,29 @@ class Interpolator(ABC):
     def get_laplacian(self, x, y, z):
         pass
     ############################################################################################################
-
     # implemented interface that is the same for all abstractions
+    def make_radient(self,dx,dy,dz):
+        radient = abs(dx) + abs(dy) + abs(dz)
+        #radient = dx+dy+dz                                
+        return radient
+    
+    def make_laplacian(self,ddx,ddy,ddz):
+        laplacian = ddx + ddy + ddz 
+        return laplacian
+
     def get_radient_numerical(self, x, y, z):                        
         val = self.get_value(x, y, z)
         dx = (self.get_value(x + self.h, y, z) - val) / self.h
         dy = (self.get_value(x, y + self.h, z) - val) / self.h
-        dz = (self.get_value(x, y, z + self.h) - val) / self.h
-        radient = (abs(dx) + abs(dy) + abs(dz)) / 3
-        return radient
+        dz = (self.get_value(x, y, z + self.h) - val) / self.h        
+        return self.make_radient(dx,dy,dz)
         
     def get_laplacian_numerical(self, x, y, z):        
         val = self.get_value(x, y, z)
         xx = self.getDxDx_numerical(x, y, z, val)
         yy = self.getDyDy_numerical(x, y, z, val)
         zz = self.getDzDz_numerical(x, y, z, val)
-        return xx + yy + zz 
+        return self.make_laplacian(xx,yy,zz)
         
     def getDxDx_numerical(self, x, y, z, val):        
         va = self.get_value(x - self.h, y, z)
@@ -94,19 +108,38 @@ class Interpolator(ABC):
         dd = (va + vb - 2 * val) / (self.h * self.h)
         return dd
                 
-    def get_fms(self,f,m,s,F=-1,M=-1,S=-1):        
-        u_f, u_m, u_s = self.get_adjusted_fms(f,m,s,F,M,S)
-        return self._npy[int(u_f),int(u_m),int(u_s)]
-        #pos = self.get_pos_from_fms(u_f,u_m,u_s,F,M,S)
-        #return self._values[pos]
-    
-    def get_projection(self,slice):        
+    def get_fms(self,f,m,s):        
+        #u_f, u_m, u_s = self.get_adjusted_fms(f,m,s)
+        return self._npy[int(f),int(m),int(s)]
+            
+    def get_projection(self,slice,xmin=-1,xmax=-1,ymin=-1,ymax=-1):
+        vals = None        
         if slice == "xy":        
-            return self._npy.max(axis=(2))
+            vals = self._npy.max(axis=(2))
         elif slice == "yz":        
-            return self._npy.max(axis=(0))
+            vals = self._npy.max(axis=(0))
         elif slice == "zx":        
-            return self._npy.max(axis=(1))
+            vals = self._npy.max(axis=(1))
+        
+        if xmin == -1 and xmax == -1 and ymin == -1 and ymax == -1:
+            return vals
+        
+        imax,jmax = vals.shape
+        newvals = np.zeros((xmax-xmin,ymax-ymin))
+        for i in range(xmin,xmax):
+            for j in range(ymin,ymax):
+                x,y = i,j
+                a,b = i-xmin,j-ymin
+                while x < 0:
+                    x += imax
+                while y < 0:
+                    y += jmax
+                x = x%imax
+                y = y%jmax
+                newvals[a,b] = vals[x,y]
+        return newvals
+        
+
     
     def get_cross_section(self,slice,layer):        
         if slice == "xy":        
@@ -126,29 +159,35 @@ class Interpolator(ABC):
         pos += f        
         return pos
     
-    def get_adjusted_fms(self, f, m, s,F=-1,M=-1,S=-1):
+    def get_adjusted_fms(self, f, m, s):
+        use_f,use_m,use_s = self._F, self._M, self._S
+        u_f = f%use_f
+        u_m = m%use_m
+        u_s = s%use_s
+        return round(u_f,self.round), round(u_m,self.round), round(u_s,self.round)
+
+    def get_adjusted_fmsx(self, f, m, s,F=-1,M=-1,S=-1):                
         use_f,use_m,use_s = self._F, self._M, self._S
         if F+M+S != -3:
-            use_f,use_m,use_s = F,M,S
-        u_f = f
-        u_m = m
-        u_s = s
+            use_f,use_m,use_s = F,M,S        
+        u_f,u_m,u_s = f,m,s
+        addon = 0        
         # Unit wrap F
-        while u_f < 0:
+        while u_f < 0+addon:
             u_f += use_f
         while u_f >= use_f:
             u_f -= use_f
         # Unit wrap M
-        while u_m < 0:
+        while u_m < 0+addon:
             u_m += use_m
         while u_m >= use_m:
             u_m -= use_m
         # Unit wrap S
-        while u_s < 0:
+        while u_s < 0+addon:
             u_s += use_s
         while u_s >= use_s:
             u_s -= use_s
-        return u_f, u_m, u_s
+        return round(u_f,self.round), round(u_m,self.round), round(u_s,self.round)
         
     def get_fms_from_pos(self, A):        
         f,m,c = 0,0,0
@@ -166,21 +205,29 @@ class Interpolator(ABC):
         return f,m,c
 
     def closest(self,point):
-        cnrs = self.corners(point)
-        dis = 10000
-        pnt = None
-        for cnr in cnrs:
-            mag = point.distance(cnr)
-            if mag < dis:
-                pnt = cnr
-                dis = mag        
-        return v3.VectorThree(pnt.A,pnt.B,pnt.C)
+        al,au = math.floor(round(point.A,self.round)),math.ceil(round(point.A,self.round))
+        bl,bu = math.floor(round(point.B,self.round)),math.ceil(round(point.B,self.round))
+        cl,cu = math.floor(round(point.C,self.round)),math.ceil(round(point.C,self.round))
+        a,b,c = 0,0,0
+        if round(point.A,self.round)-al <= au-round(point.A,self.round):
+            a=al
+        else:
+            a=au
+        if round(point.B,self.round)-bl <= bu-round(point.B,self.round):
+            b=bl
+        else:
+            b=bu
+        if round(point.C,self.round)-cl <= cu-round(point.C,self.round):
+            c=cl
+        else:
+            c=cu        
+        return v3.VectorThree(a,b,c)
     
     def corners(self,point, far=1):
         cnrs = []        
-        x = np.ceil(point.A)
-        y = np.ceil(point.B)
-        z = np.ceil(point.C)
+        x = np.ceil(round(point.A,self.round))
+        y = np.ceil(round(point.B,self.round))
+        z = np.ceil(round(point.C,self.round))
         for f in range(0-far,far):
             for m in range(0-far,far):
                 for s in range(0-far,far):
@@ -188,11 +235,11 @@ class Interpolator(ABC):
         return cnrs
     
     def get_val_slice(self,unit_coords, deriv = 0):
-        vals = []
-        for i in range(len(unit_coords)):
-            row = []
+        unit_coords = np.array(unit_coords)
+        vals = np.zeros(unit_coords.shape)
+        for i in range(len(unit_coords)):            
             for j in range(len(unit_coords[0])):
-                vec = unit_coords[i][j]
+                vec = unit_coords[i,j]
                 if self.log_level > 2:
                     print("Get value", vec.A,vec.B,vec.C)
                 if deriv == 2:
@@ -201,25 +248,49 @@ class Interpolator(ABC):
                     vec_val = self.get_radient(vec.A,vec.B,vec.C)
                 else:
                     vec_val = self.get_value(vec.A,vec.B,vec.C)
-                row.append(vec_val)
-            vals.append(row)
-        return vals
+                vals[i,j] = vec_val            
+        return vals.tolist()
 
     def build_cube_around(self, x, y, z, width):        
         # 1. Build the points around the centre as a cube - width points
-        vals = []            
+        #vals = []            
+        #print("---",x,y,z)
+        arr = np.zeros((width*width*width))        
         xp,yp,zp = 0,0,0 
+        count = 0
         for i in range(int(-1*width/2 + 1), int(width/2 + 1)):
-            xp = np.floor(x + i)
+            xp = np.floor(round(x,self.round)+i)
             for j in range(int(-1*width/2 + 1), int(width/2 + 1)):      
-                yp = np.floor(y + j)
+                yp = np.floor(round(y,self.round)+j)
                 for k in range(int(-1*width/2 + 1), int(width/2 + 1)):          
-                    zp = np.floor(z + k)                        
-                    p = self.get_fms(xp, yp, zp)
-                    vals.append(p)                    
-        npvals = np.array(vals)    
-        return npvals
+                    zp = np.floor(round(z,self.round)+k)
+                    xp_u, yp_u, zp_u = self.get_adjusted_fms(xp,yp,zp)
+                    p = self.get_fms(xp_u, yp_u, zp_u)
+                    #print(xp,yp,zp)
+                    #vals.append(p)                    
+                    arr[count] = p
+                    count += 1
+        #npvals = np.array(vals)    
+        #return npvals
+        return arr
 
+    def build_cube_aroundx(self, x, y, z, width):        
+        # 1. Build the points around the centre as a cube - width points        
+        arr = np.zeros((width*width*width))        
+        xp,yp,zp = 0,0,0         
+        half = int(width/2)        
+        count = 0
+        for i in range(width):
+            xp = np.floor(round(x,self.round))+i-half-1
+            for j in range(width):
+                yp = np.floor(round(y,self.round))+j-half-1
+                for k in range(width):
+                    zp = np.floor(round(z,self.round))+k-half-1
+                    p = self.get_fms(xp, yp, zp)
+                    arr[count] = p
+                    count += 1
+        return arr
+    
     def mult_vector(self,A, V):     #A and V are numpy arrays
         use_np = True
         if use_np:
@@ -244,11 +315,12 @@ class Interpolator(ABC):
 class Nearest(Interpolator):                
     def init(self):
         pass
-    def get_value(self, x, y, z):
-        closest_pnt = self.closest(v3.VectorThree(x,y,z))  
-        #print(closest_pnt.A, closest_pnt.B,closest_pnt.C)      
-        return self.get_fms(closest_pnt.A, closest_pnt.B,closest_pnt.C)    
     ## implement abstract interface #########################################
+    def get_value(self, x, y, z):
+        closest_pnt = self.closest(v3.VectorThree(x,y,z))
+        u_f, u_m, u_s = self.get_adjusted_fms(closest_pnt.A, closest_pnt.B,closest_pnt.C)        
+        return self.get_fms(u_f, u_m, u_s)
+    
     def get_radient(self, x, y, z):
         return self.get_radient_numerical(x,y,z)
     
@@ -275,27 +347,36 @@ class Multivariate(Interpolator):
     
     ## implement abstract interface #########################################
     def get_value(self, x, y, z):
-        return self.get_value_new(x,y,z)
+        u_x, u_y, u_z = self.get_adjusted_fms(x,y,z)
+        xn,yn,zn,coeffs = self.make_coeffs(u_x,u_y,u_z)
+        return self.get_value_multivariate(zn, yn, xn, coeffs)
 
+    def get_radientx(self, x, y, z):
+        return self.get_radient_numerical(x,y,z)
+    
+    def get_laplacianx(self, x, y, z):
+        return self.get_laplacian_numerical(x,y,z)
+    
     def get_radient(self, x, y, z):
-        xn,yn,zn,coeffs = self.make_coeffs(x,y,z)
+        u_x, u_y, u_z = self.get_adjusted_fms(x,y,z)
+        xn,yn,zn,coeffs = self.make_coeffs(u_x,u_y,u_z)
         dx = self.get_value_multivariate(zn, yn, xn, coeffs,["x"])
         dy = self.get_value_multivariate(zn, yn, xn, coeffs,["y"])
         dz = self.get_value_multivariate(zn, yn, xn, coeffs,["z"])
-        return (abs(dx) + abs(dy) + abs(dz))/3
-    
+        return self.make_radient(dx,dy,dz)
+            
     def get_laplacian(self, x, y, z):
-        xn,yn,zn,coeffs = self.make_coeffs(x,y,z)
+        u_x, u_y, u_z = self.get_adjusted_fms(x,y,z)
+        xn,yn,zn,coeffs = self.make_coeffs(u_x,u_y,u_z)
         ddx = self.get_value_multivariate(zn, yn, xn, coeffs,["x","x"])
         ddy = self.get_value_multivariate(zn, yn, xn, coeffs,["y","y"])
         ddz = self.get_value_multivariate(zn, yn, xn, coeffs,["z","z"])
-        return ddx+ddy+ddz
+        return self.make_laplacian(ddx,ddy,ddz)
+        
+    
     ## iplement abstract interface ###########################################
-
-    def get_value_new(self, x, y, z):
-        xn,yn,zn,coeffs = self.make_coeffs(x,y,z)
-        return self.get_value_multivariate(zn, yn, xn, coeffs)
-
+    
+    """
     def get_value_old(self, x, y, z):
         # The method of linear interpolation is a version of my own method for multivariate fitting, instead of trilinear interpolation
         # NOTE I could extend this to be multivariate not linear but it has no advantage over bspline - and is slower and not as good 
@@ -339,11 +420,12 @@ class Multivariate(Interpolator):
         #5. Apply the multivariate polynomial coefficents to find the value
         #return self.get_value_multivariate(zn, yn, xn, self.polyCoeffs)
         return self.get_value_multivariate(zn, yn, xn, self.polyCoeffs)
+    """
         
     def get_value_multivariate(self, x, y, z, coeffs,wrt=[]):        
         #This is using a value scheme that makes sens of our new fitted polyCube
         #In a linear case it will be a decimal between 0 and 1          
-        calc_mat = coeffs
+        calc_mat = np.copy(coeffs)
         for partial in wrt:
             if partial == "x":
                 calc_mat = self.diffWRTx(calc_mat)
@@ -364,7 +446,7 @@ class Multivariate(Interpolator):
 
     def diffWRTx(self,coeffs):        
         x,y,z = coeffs.shape
-        partialX = np.zeros((x,y,z))
+        partialX = np.zeros((x-1,y,z))
         for i in range(1, x):
             for j in range(0,y):
                 for k in range(0,z):
@@ -373,7 +455,7 @@ class Multivariate(Interpolator):
 
     def diffWRTy(self,coeffs):        
         x,y,z = coeffs.shape
-        partialY = np.zeros((x,y,z))
+        partialY = np.zeros((x,y-1,z))
         for i in range(0, x):
             for j in range(1,y):
                 for k in range(0,z):
@@ -382,7 +464,7 @@ class Multivariate(Interpolator):
     
     def diffWRTz(self,coeffs):        
         x,y,z = coeffs.shape
-        partialZ = np.zeros((x,y,z))
+        partialZ = np.zeros((x,y,z-1))
         for i in range(0, x):
             for j in range(0,y):
                 for k in range(1,z):
@@ -393,21 +475,34 @@ class Multivariate(Interpolator):
         # 1. Build the points around the centre as a cube - 8 points
         vals = self.build_cube_around(x, y, z, self.points)
         #2. Multiply with the precomputed matrix to find the multivariate polynomial
-        ABC = self.mult_vector(self.inv.get_invariant(), vals)
-        # 3. Put the 8 values back into a cube
-        polyCoeffs = np.zeros((self.points, self.points, self.points))
-        pos = 0
-        for i in range(self.points):
-            for j in range(self.points):                
-                for k in range(self.points):                    
-                    polyCoeffs[i, j, k] = ABC[pos]
-                    pos+=1        
-        #4. Adjust the values to be within this cube
+        polyCoeffs = self.mult_vector(self.inv.get_invariant(), vals)
+        polyCoeffs = polyCoeffs.reshape((self.points,self.points,self.points))
+        #ABC = self.mult_vector(self.inv.get_invariant(), vals)        
+        # 3. Put the 8 values back into a cube                
+        #polyCoeffs = np.zeros((self.points, self.points, self.points))
+        #pos = 0
+        #for i in range(self.points):
+        #    for j in range(self.points):                
+        #        for k in range(self.points):                    
+        #            polyCoeffs[i, j, k] = ABC[pos]
+        #            pos+=1        
+        
+        #4. Adjust the values to be within this cube        
         pstart = (-1 * self.points / 2) + 1        
-        xn = x - np.floor(x) - pstart
-        yn = y - np.floor(y) - pstart
-        zn = z - np.floor(z) - pstart
+        xn = round(x - np.floor(x) - pstart,self.round)
+        yn = round(y - np.floor(y) - pstart,self.round)
+        zn = round(z - np.floor(z) - pstart,self.round)
+        if (xn > 1 or yn > 1 or zn > 1) and self.points==2:
+            print(xn,yn,zn,x,y,z,self.points)
+        elif xn < 0 or yn < 0 or zn < 0 and self.points==2:
+            print(xn,yn,zn,x,y,z,self.points)
+        elif (xn > 2 or yn > 2 or zn > 2) and self.points==4:
+            print(xn,yn,zn,x,y,z,self.points)
+        elif (xn < 1 or yn < 1 or zn < 1) and self.points==4:
+            print(xn,yn,zn,x,y,z,self.points)
+        
         return xn,yn,zn,polyCoeffs
+    
             
 ####################################################################################################
 ### B-Spline
@@ -419,11 +514,11 @@ class Bspline(Interpolator):
     http://bigwww.epfl.ch/thevenaz/interpolation/
     *******************************************************************************
     """
-    def init(self):        
-        self.TOLERANCE = 2.2204460492503131e-016 # smallest such that 1.0+DBL_EPSILON != 1.0                
-        self.mirror = False                
-        self.make_periodic_coeffs() #temp make it just ordinary coeffs
-            
+    def init(self):                
+        self.padded = True
+        self.make_periodic_coeffs() 
+
+                    
     ## implement abstract interface #########################################
     def get_radient(self, x, y, z):
         return self.get_radient_numerical(x,y,z)
@@ -432,14 +527,25 @@ class Bspline(Interpolator):
         return self.get_laplacian_numerical(x,y,z)
     ## iplement abstract interface ###########################################
 
-    def make_periodic_coeffs(self):           
-        # we make the coefficients matrix a bit bigger than the vaues and have it wrap, and then cut it back down to the values                 
-        self._coeffs =self.extend_vals_with_buffer(self._buffer)
-        tmpF = self._F + (2 * self._buffer)
-        tmpM = self._M + (2 * self._buffer)
-        tmpS = self._S + (2 * self._buffer)                
-        self.create_coeffs(tmpF,tmpM,tmpS)        
-        self._coeffs = self.reduce_vals_with_buffer(self._buffer, self._coeffs)
+    def simple_copy_coeffs(self):      
+        self._coeffs = np.copy(self._npy)
+
+    def make_periodic_coeffs(self):
+        # we make the coefficients matrix a bit bigger than the vaues and have it wrap, and then cut it back down to the values                         
+        self.padded = False
+        if not self.padded:
+            self.simple_copy_coeffs()            
+            #self.create_coeffs_copy(self._F,self._M,self._S)                    
+            self.create_coeffs_inplace(self._F,self._M,self._S)                    
+        else:
+            self._coeffs =self.extend_vals_with_buffer(self.buffer)
+            #print(self._coeffs)            
+            #self.create_coeffs_copy(self._F+2*self.buffer,self._M+2*self.buffer,self._S+2*self.buffer)        
+            self.create_coeffs_inplace(self._F+2*self.buffer,self._M+2*self.buffer,self._S+2*self.buffer)        
+            #print(self._coeffs)
+            #self._coeffs = self.reduce_vals_with_buffer(self.buffer, self._coeffs)
+        #print(self._npy)
+        #print(self._coeffs)
                 
     def reduce_vals_with_buffer(self,buffer, mynpy):        
         xx,yy,zz = mynpy.shape
@@ -450,47 +556,95 @@ class Bspline(Interpolator):
         #// 1. Make a buffer padded values cube for periodic values        
         xxa,yya,zza = self._F, self._M, self._S        
         padded = np.pad(self._npy,[(buffer,buffer),(buffer,buffer),(buffer,buffer)])        
+        # pad z
         for z in range(buffer):
-            for y in range(buffer):
-                for x in range(buffer):
-                    p = self.get_fms(x-buffer,y-buffer,z-buffer)
+            for y in range(-1*buffer, self._M+buffer):
+                for x in range(-1*buffer, self._F+buffer):
+                    xe,ye,ze = self.get_adjusted_fms(x-buffer,y-buffer,z-buffer)
+                    p = self.get_fms(xe,ye,ze)
                     padded[x,y,z] = p                            
         for z in range(zza,zza+buffer):
-            for y in range(yya,yya+buffer):
-                for x in range(xxa,xxa+buffer):
-                    p = self.get_fms(x-buffer,y-buffer,z-buffer)                    
+            for y in range(-1*buffer, self._M+buffer):
+                for x in range(-1*buffer, self._F+buffer):
+                    xe,ye,ze = self.get_adjusted_fms(x-buffer,y-buffer,z-buffer)
+                    p = self.get_fms(xe,ye,ze)
+                    padded[x,y,z] = p                                            
+        # pad y
+        for y in range(buffer):
+            for z in range(-1*buffer, self._S+buffer):
+                for x in range(-1*buffer, self._F+buffer):
+                    xe,ye,ze = self.get_adjusted_fms(x-buffer,y-buffer,z-buffer)
+                    p = self.get_fms(xe,ye,ze)
+                    padded[x,y,z] = p                            
+        for y in range(yya,yya+buffer):
+            for z in range(-1*buffer, self._S+buffer):
+                for x in range(-1*buffer, self._F+buffer):
+                    xe,ye,ze = self.get_adjusted_fms(x-buffer,y-buffer,z-buffer)
+                    p = self.get_fms(xe,ye,ze)
                     padded[x,y,z] = p                                    
+        # pad x
+        for x in range(buffer):
+            for z in range(-1*buffer, self._S+buffer):
+                for y in range(-1*buffer, self._M+buffer):
+                    xe,ye,ze = self.get_adjusted_fms(x-buffer,y-buffer,z-buffer)
+                    p = self.get_fms(xe,ye,ze)
+                    padded[x,y,z] = p                            
+        for x in range(xxa,xxa+buffer):
+            for z in range(-1*buffer, self._S+buffer):
+                for y in range(-1*buffer, self._M+buffer):
+                    xe,ye,ze = self.get_adjusted_fms(x-buffer,y-buffer,z-buffer)
+                    p = self.get_fms(xe,ye,ze)
+                    padded[x,y,z] = p                                    
+        
+        
         return padded
                                                                   
-    def create_coeffs(self, thisF, thisM, thisS):
+    def create_coeffs_inplace(self, thisF, thisM, thisS):
+        pole = self.get_pole(self.degree)
+        num_poles = len(pole)
+        #Convert the samples to interpolation coefficients
+        #X-wise
+        for y in range(thisM):        
+            for z in range(thisS):                        
+                row = self._coeffs[:,y:y+1,z:z+1].reshape([thisF])
+                #self.convert_to_interp_coeffs_inplace(pole, num_poles, thisF, row)
+                self.convert_to_interp_coeffs(pole, num_poles, thisF, row)
+        #Y-wise
+        for x in range(thisF):        
+            for z in range(thisS):                        
+                row = self._coeffs[x:x+1,:,z:z+1].reshape([thisM])
+                #self.convert_to_interp_coeffs_inplace(pole, num_poles, thisM, row)
+                self.convert_to_interp_coeffs(pole, num_poles, thisM, row)
+        #Z-wise
+        for x in range(thisF):
+            for y in range(thisM):                        
+                row = self._coeffs[x:x+1,y:y+1,:].reshape([thisS])
+                #line = self.convert_to_interp_coeffs_inplace(pole, num_poles, thisS, row)
+                line = self.convert_to_interp_coeffs(pole, num_poles, thisS, row)
+
+    def create_coeffs_copy(self, thisF, thisM, thisS):
         pole = self.get_pole(self.degree)
         num_poles = len(pole)
         #Convert the samples to interpolation coefficients
         #X-wise
         for y in range(thisM):        
             for z in range(thisS):        
-                #row = self.get_row3d(y, z, thisF,thisF, thisM, thisS)
-                #line = self.convert_to_interp_coeffs(pole, num_poles, thisF, row,self.mirror)
-                #self.put_row3d(y, z, line, thisF,thisF, thisM, thisS)
-                row = self._coeffs[:,y:y+1,z:z+1].reshape([thisF])
-                self.convert_to_interp_coeffs(pole, num_poles, thisF, row,self.mirror)
+                row = self.get_row3d(y, z, thisF)
+                line = self.convert_to_interp_coeffs(pole, num_poles, thisF, row)
+                self.put_row3d(y, z, line, thisF)
         #Y-wise
         for x in range(thisF):        
             for z in range(thisS):        
-                #row = self.get_col3d(x, z, thisM,thisF, thisM, thisS)
-                #line = self.convert_to_interp_coeffs(pole, num_poles, thisM, row,self.mirror)
-                #self.put_col3d(x, z, line, thisM,thisF, thisM, thisS)           
-                row = self._coeffs[x:x+1,:,z:z+1].reshape([thisM])
-                self.convert_to_interp_coeffs(pole, num_poles, thisM, row,self.mirror)
+                row = self.get_col3d(x, z, thisM)
+                line = self.convert_to_interp_coeffs(pole, num_poles, thisM, row)
+                self.put_col3d(x, z, line, thisM)
         #Z-wise
         for x in range(thisF):
             for y in range(thisM):        
-                #row = self.get_hole3d(x, y, thisS,thisF, thisM, thisS)
-                #line = self.convert_to_interp_coeffs(pole, num_poles, thisS, row,self.mirror)
-                #self.put_hole3d(x, y, line, thisS,thisF, thisM, thisS)
-                row = self._coeffs[x:x+1,y:y+1,:].reshape([thisS])
-                line = self.convert_to_interp_coeffs(pole, num_poles, thisS, row,self.mirror)
-                                            
+                row = self.get_hole3d(x, y, thisS)
+                line = self.convert_to_interp_coeffs(pole, num_poles, thisS, row)
+                self.put_hole3d(x, y, line, thisS)
+
     def get_pole(self,degree):        
         #Recover the poles from a lookup table #currently only 3 degree, will I want to calculate all the possibilities at the beginnning, 3,5,7,9?
         pole = []
@@ -510,37 +664,37 @@ class Bspline(Interpolator):
             pole.append(np.sqrt(3.0) - 2.0)
         return pole
 
-    def get_row3d(self,y,z,length,F,M,S):        
+    def get_row3d(self,y,z,length):        
         row = []
         for x in range(length):
-            row.append(self.get_coeff(x, y, z,F,M,S))
+            row.append(self.get_coeff(x, y, z))
         return row
     
-    def get_col3d(self,x,z,length,F,M,S):        
+    def get_col3d(self,x,z,length):        
         col = []
         for y in range(length):
-            col.append(self.get_coeff(x, y, z,F,M,S))
+            col.append(self.get_coeff(x, y, z))
         return col
     
-    def get_hole3d(self,x,y,length,F,M,S):        
+    def get_hole3d(self,x,y,length):        
         bore = []
         for z in range(length):
-            bore.append(self.get_coeff(x, y, z,F,M,S))
+            bore.append(self.get_coeff(x,y,z))
         return bore
     
-    def put_row3d(self,y,z,row,length,F,M,S):        
+    def put_row3d(self,y,z,row,length):        
         for x in range(length):
-            self.put_coeff(x, y, z, row[x],F,M,S)
+            self.put_coeff(x, y, z, row[x])
     
-    def put_col3d(self,x,z,col,length,F,M,S):        
+    def put_col3d(self,x,z,col,length):        
         for y in range(length):
-            self.put_coeff(x, y, z, col[y],F,M,S)
+            self.put_coeff(x, y, z, col[y])
     
-    def put_hole3d(self,x,y,bore,length,F,M,S):        
+    def put_hole3d(self,x,y,bore,length):        
         for z in range(length):
-            self.put_coeff(x, y, z, bore[z],F,M,S)
+            self.put_coeff(x, y, z, bore[z])
     
-    def convert_to_interp_coeffs(self,pole, num_poles, width, row,mirror):        
+    def convert_to_interp_coeffs(self,pole, num_poles, width, row):        
         #/* special case required by mirror boundaries */
         if (width == 1):         
             #mirror filter and periodic filter
@@ -557,23 +711,47 @@ class Bspline(Interpolator):
             row[n] = row[n] * lmbda        
         #loop over the poles            
         for k in range(num_poles):        
-            #/* causal initialization */
-            if (mirror):
-                row[0] = self.initial_causal_coeffs(row, width, pole[k])
-            else:# //TODO                       
-                row[0] = self.initial_causal_coeffs(row, width, pole[k])            
+            #/* causal initialization */            
+            row[0] = self.initial_causal_coeffs(row, width, pole[k])            
             #/* causal recursion */
             for n in range(1,width):            
                 row[n] = row[n]+ pole[k] * row[n - 1]
-            #/* anticausal initialization */
-            if (mirror):
-                row[width - 1] = self.initial_anticausal_coeffs(row, width, pole[k])
-            else:# //TODO
-                row[width - 1] = self.initial_anticausal_coeffs(row, width, pole[k])
+            #/* anticausal initialization */                        
+            row[width - 1] = self.initial_anticausal_coeffs(row, width, pole[k])
             #/* anticausal recursion */
             for n in range(width - 2, -1, -1):
                 row[n] = pole[k] * (row[n + 1] - row[n])
         return row
+
+    def convert_to_interp_coeffs_inplace(self,pole, num_poles, width, row):        
+        #/* This version is periodic ONLY */
+        if (width == 1):#not much can be done if it is only 1 think            
+            return row        
+        lmbda = 1
+        #Compute the overall gain
+        for k in range(num_poles):
+            lmbda = lmbda * (1 - pole[k]) * (1 - 1 / pole[k])        
+        #Apply the gain
+        for n in range(width):
+            row[n] = row[n] * lmbda        
+        #loop over the poles            
+        for k in range(num_poles):                                
+            #/* causal recursion */            
+            for n in range(0,width):
+                m = n
+                if m < 0:
+                    m = -1*width%abs(n)
+                n1 = m-1                
+                row[m] = row[m]+ pole[k] * row[n1]                                    
+            
+            #/* anticausal recursion */            
+            for n in range(width-1, -1, -1):
+                n1 = n+1
+                if n1 >= width:
+                    n1 = n1%width
+                m = n1-1
+                row[m] = pole[k] * (row[n1] - row[m])
+        return row
     
     def initial_causal_coeffs(self, vals, length, pole):
         #/* begin InitialCausalCoefficient */
@@ -592,14 +770,14 @@ class Bspline(Interpolator):
                 zn *= pole       
             return Sum        
         else:#// if (_mirror)
-            #// RSA notes this is the mirror condition, when the horizon - how far you need to look ahead for good data - is not as far the data you have
+            #// RSA notes is this a mirror condition? when the horizon - how far you need to look ahead for good data - is not as far the data you have
             #/* full loop */
             zn = pole
             iz = 1.0 / pole
             z2n = math.pow(pole, length - 1)
             Sum = vals[0] + z2n * vals[length - 1]
             z2n *= z2n * iz 
-            for n in range(1,length - 2):
+            for n in range(1,length - 1):
                 Sum += (zn + z2n) * vals[n]
                 zn *= pole                
                 z2n *= iz
@@ -612,20 +790,20 @@ class Bspline(Interpolator):
         else:#// if (_mirror)
             return ((pole / (pole * pole - 1.0)) * (pole * vals[length - 2] + vals[length - 1]))
 
-    def get_coeff(self,x,y,z,F,M,S):        
+    def get_coeff(self,x,y,z):        
         # this naturally wraps round
-        u_x, u_y, u_z = self.get_adjusted_fms(x,y,z,F=F,M=M,S=S)
-        #pos = int(self.get_pos_from_fms(u_x,u_y,u_z,F=F,M=M,S=S))
-        value = self._coeffs[u_x,u_y,u_z]
+        #u_x, u_y, u_z = self.get_adjusted_fms(x,y,z)        
+        value = self._coeffs[x,y,z]
         return value
                                         
-    def put_coeff(self,x,y,z,v,F,M,S):
-        #u_x, u_y, u_z = self.get_adjusted_fms(x,y,z,F=F,M=M,S=S)
-        #pos = int(self.get_pos_from_fms(u_x,u_y,u_z,F=F,M=M,S=S))
+    def put_coeff(self,x,y,z,v):        
         self._coeffs[x,y,z] = v
 
-    def get_value(self, x, y, z):        
-        u_x, u_y, u_z = self.get_adjusted_fms(x,y,z)
+    def get_value(self, x, y, z):
+        u_x, u_y, u_z = x,y,z        
+        if self.padded:
+            u_x, u_y, u_z = u_x+self.buffer,u_y+self.buffer,u_z+self.buffer        
+        u_x, u_y, u_z = self.get_adjusted_fms(u_x, u_y, u_z)
         weight_length = self.degree + 1
         xIndex, yIndex, zIndex = [],[],[]        
         xWeight, yWeight,zWeight = [],[],[]        
@@ -637,9 +815,9 @@ class Bspline(Interpolator):
             yWeight.append(0)
             zWeight.append(0)            
         #Compute the interpolation indices
-        i = int(np.floor(u_x) - np.floor(self.degree / 2))
-        j = int(np.floor(u_y) - np.floor(self.degree / 2))
-        k = int(np.floor(u_z) - np.floor(self.degree / 2))
+        i = int(np.floor(round(u_x,self.round)) - np.floor(self.degree / 2))
+        j = int(np.floor(round(u_y,self.round)) - np.floor(self.degree / 2))
+        k = int(np.floor(round(u_z,self.round)) - np.floor(self.degree / 2))
 
         for l in range(self.degree+1):                          
             xIndex[l] = i #if 71.1 passed in, for linear, we would want 71 and 72, 
@@ -673,8 +851,13 @@ class Bspline(Interpolator):
             w2 = 0.0
             for j in range(spline_degree+1):
                 w1 = 0.0
-                for i in range(spline_degree+1):                
-                    w1 += xWeight[i] * self.get_coeff(xIndex[i], yIndex[j], zIndex[k],self._F,self._M,self._S)
+                for i in range(spline_degree+1):
+                    xc,yc,zc = xIndex[i],yIndex[j],zIndex[k]
+                    xc,yc,zc = self.get_adjusted_fms(xc,yc,zc)
+                    cc = self.get_coeff(xc,yc,zc)
+                    #print(xIndex[i],yIndex[j],zIndex[k])
+                    #print(xc,yc,zc,cc)                    
+                    w1 += xWeight[i] * cc
                 w2 += yWeight[j] * w1
             w3 += zWeight[k] * w2
         return w3
@@ -684,10 +867,20 @@ class Bspline(Interpolator):
         for i in range(weight_length):
             ws.append(0)
         w = val - idc[1]
+        if w > 1:
+            print("w=",w)
+        if w < 0:
+            print("w=",w)
+        # Making it linear
+        #ws[3] = 1*w/40
+        #ws[0] = 1*(1-w) / 40
+        #ws[2] = 39*w/40
+        #ws[1] = 39*(1-w) / 40        
         ws[3] = (1.0 / 6.0) * w * w * w
-        ws[0] = (1.0 / 6.0) + (1.0 / 2.0) * w * (w - 1.0) - ws[3]
+        ws[0] = (1.0 / 6.0) + (1.0 / 2.0) * w * (w - 1.0) - ws[3]        
         ws[2] = w + ws[0] - 2.0 * ws[3]
         ws[1] = 1.0 - ws[0] - ws[2] - ws[3]
+        
         return ws
         
     def applyValue5(self,val,idc,weight_length):        
@@ -755,7 +948,20 @@ class Bspline(Interpolator):
         ws[8] = 1.0 - ws[0] - ws[1] - ws[2] - ws[3] - ws[4] - ws[5] - ws[6] - ws[7] - ws[9]
         return ws
         
-        
+####################################################################################################
+### R-Spline
+#################################################################################################### 
+class Mspline(Bspline):
+    def init(self):                
+        #self.make_periodic_coeffs(True) 
+        #self.make_periodic_coeffs(False) 
+        self.padded = False
+        self.make_periodic_coeffs()
+        #self.simple_copy_coeffs() #temp make it just ordinary coeffs
+    
+    
+
+    
             
         
         
